@@ -52,6 +52,10 @@
 #include "global.h"
 
 #include <stdio.h>
+#ifdef HAVE_READLINE
+# include <readline/readline.h>
+# include <readline/history.h>
+#endif
 #include <stdlib.h>
 #include <unistd.h>
 #include <string.h>
@@ -72,9 +76,6 @@
 #include "romio.h"
 #include "resources.h"
 
-extern char *readline __ProtoType__ ((const char *));
-extern void add_history __ProtoType__ ((char *));
-
 #define MAX_ARGS 16
 
 int enter_debugger = 0;
@@ -89,6 +90,9 @@ static char instr[100];
  */
 #define DSKTOP_SX	0x70579
 #define DSKBOT_SX	0x7057e
+/*
+ * Pointers in the HP48gx ROM
+ */
 #define DSKTOP_GX	0x806f8
 #define DSKBOT_GX	0x806fd
 
@@ -344,7 +348,6 @@ check_breakpoint (type, addr)
   return 0;
 }
 
-
 char *
 #ifdef __FunctionProto__
 read_str(char *str, int n, int fp)
@@ -362,7 +365,10 @@ read_str(str, n, fp)
     {
       cc = read(fp, str, n);
       if (cc > 0)
-        return str;
+        {
+	  str[cc] = '\0';
+          return str;
+	}
       if (cc == 0)
         return NULL;
 
@@ -376,11 +382,52 @@ read_str(str, n, fp)
           fcntl(fp, F_SETFL, flags);
           continue;
         }
- 
       return NULL;
     }
+    /* not reached */
 }
 
+#ifndef HAVE_READLINE
+char *
+#ifdef __FunctionProto__
+readline(char *prompt)
+#else
+readline(prompt)
+	char *prompt;
+#endif
+{
+  char rl[81];
+  char *nrl;
+  if (prompt)
+    {
+      printf("%s", prompt);
+      fflush(stdout);
+    }
+  nrl = read_str(rl, 80, 0);
+  if (nrl)
+    {
+      int len;
+      char *str;
+
+      len = strlen(nrl);
+      str = malloc(len + 1);
+
+      if (nrl[len-1] == '\n')
+        {
+	  nrl[len-1] = '\0';
+	}
+      if (NULL == str)
+        {
+	  fprintf(stderr, "Out of memory\n");
+	  exit(-1);
+        }
+
+      strcpy(str, nrl);
+      nrl = str;
+    }
+  return nrl;
+}
+#endif
 
 static inline void
 #ifdef __FunctionProto__
@@ -531,7 +578,27 @@ decode_64 (addr, arg)
       printf ("Command requires an argument.\n");
       return 0;
     }
-
+#ifdef SIMPLE_64
+  *addr = 0;
+  for (i = 0; i < strlen (arg); i++)
+    {
+      *addr <<= 4;
+      if ('0' <= arg[i] && arg[i] <= '9')
+	{
+	  *addr |= ((int) arg[i] - (int) '0');
+	}
+      else if ('A' <= arg[i] && arg[i] <= 'F')
+	{
+	  *addr |= ((int) arg[i] - (int) 'A' + 10);
+	}
+      else
+	{
+	  *addr = 0;
+	  printf ("Not a number: %s.\n", arg);
+	  return 0;
+	}
+    }
+#else
   addr->lo = addr->hi = 0;
   for (i = 0; i < strlen (arg); i++)
     {
@@ -553,6 +620,7 @@ decode_64 (addr, arg)
 	  return 0;
 	}
     }
+#endif
   return 1;
 }
 
@@ -955,10 +1023,11 @@ set_reg (val, n, r)
 
   for (i = 0; i < n; i++)
     {
-      if (i < 8)
-	r[i] = (unsigned char) ((val.lo & (0xf << (4 * i))) >> (4 * i));
-      else
-	r[i] = (unsigned char) ((val.hi & (0xf << (4 * (i - 8)))) >> (4 * (i - 8)));
+#ifdef SIMPLE_64
+      r[i] = (unsigned char) ((val & (0xf << (4 * i))) >> (4 * i));
+#else
+      r[i] = (unsigned char) (((i < 8?val.lo:val.hi) & (0xf << (4 * i))) >> (4 * i));
+#endif
     }
 }
 
@@ -994,7 +1063,11 @@ set_st (val)
   int i;
 
   for (i = 0; i < 16; i++)
+#ifdef SIMPLE_64
+    saturn.PSTAT[i] = (val & (1 << i)) ? 1 : 0;
+#else
     saturn.PSTAT[i] = (val.lo & (1 << i)) ? 1 : 0;
+#endif
 }
 
 static void
@@ -1047,6 +1120,16 @@ set_hst (val)
   saturn.SB = 0;
   saturn.SR = 0;
   saturn.MP = 0;
+#ifdef SIMPLE_64
+  if (val & 1)
+    saturn.XM = 1;
+  if (val & 2)
+    saturn.SB = 1;
+  if (val & 4)
+    saturn.SR = 1;
+  if (val & 8)
+    saturn.MP = 1;
+#else
   if (val.lo & 1)
     saturn.XM = 1;
   if (val.lo & 2)
@@ -1055,6 +1138,7 @@ set_hst (val)
     saturn.SR = 1;
   if (val.lo & 8)
     saturn.MP = 1;
+#endif
 }
 
 static void
@@ -1079,7 +1163,7 @@ dump_hst ()
 }
 
 static char *mctl_str_gx[] = {
-  "MMIO       ", 
+  "MMIO       ",
   "SysRAM     ",
   "Bank Switch",
   "Port 1     ",
@@ -1088,7 +1172,7 @@ static char *mctl_str_gx[] = {
 };
 
 static char *mctl_str_sx[] = {
-  "MMIO  ", 
+  "MMIO  ",
   "SysRAM",
   "Port 1",
   "Port 2",
@@ -1303,7 +1387,11 @@ do_regs (argc, argv)
 	    }
 	  else if (!strcmp ("D0", argv[1]))
 	    {
+#ifdef SIMPLE_64
+	      saturn.D0 = (word_20)(val & 0xfffff);
+#else
 	      saturn.D0 = (word_20)(val.lo & 0xfffff);
+#endif
 	      printf ("    D0:\t%.5lX ->", saturn.D0);
 	      for (i = 0; i < 20; i += 5)
 		{
@@ -1313,7 +1401,11 @@ do_regs (argc, argv)
 	    }
 	  else if (!strcmp ("D1", argv[1]))
 	    {
+#ifdef SIMPLE_64
+	      saturn.D1 = (word_20)(val & 0xfffff);
+#else
 	      saturn.D1 = (word_20)(val.lo & 0xfffff);
+#endif
 	      printf ("    D1:\t%.5lX ->", saturn.D1);
 	      for (i = 0; i < 20; i += 5)
 		{
@@ -1323,12 +1415,20 @@ do_regs (argc, argv)
 	    }
 	  else if (!strcmp ("P", argv[1]))
 	    {
+#ifdef SIMPLE_64
+	      saturn.P = (word_4)(val & 0xf);
+#else
 	      saturn.P = (word_4)(val.lo & 0xf);
+#endif
 	      printf ("     P:\t%.1X\n", saturn.P);
 	    }
 	  else if (!strcmp ("PC", argv[1]))
 	    {
+#ifdef SIMPLE_64
+	      saturn.PC = (word_20)(val & 0xfffff);
+#else
 	      saturn.PC = (word_20)(val.lo & 0xfffff);
+#endif
 	      disassemble (saturn.PC, instr);
 	      printf ("    PC:\t%.5lX -> %s\n", saturn.PC, instr);
 	    }
@@ -1369,12 +1469,20 @@ do_regs (argc, argv)
 	    }
 	  else if (!strcmp ("CARRY", argv[1]))
 	    {
+#ifdef SIMPLE_64
+	      saturn.CARRY = (word_1)(val & 0x1);
+#else
 	      saturn.CARRY = (word_1)(val.lo & 0x1);
+#endif
 	      printf (" CARRY:\t%.1d\n", saturn.CARRY);
 	    }
 	  else if (!strcmp ("CY", argv[1]))
 	    {
+#ifdef SIMPLE_64
+	      saturn.CARRY = (word_1)(val & 0x1);
+#else
 	      saturn.CARRY = (word_1)(val.lo & 0x1);
+#endif
 	      printf (" CARRY:\t%.1d\n", saturn.CARRY);
 	    }
 	  else if (!strcmp ("ST", argv[1]))
@@ -1420,6 +1528,66 @@ struct se {
   struct se *se_next;
 };
 
+char *
+#ifdef __FunctionProto__
+get_stack (void)
+#else
+get_stack ()
+#endif
+{
+  word_20 dsktop, dskbot;
+  word_20 sp = 0, end = 0, ent = 0;
+  word_20 ram_base, ram_mask;
+  char    dat[65536];
+  char    typ[256];
+  int i, n;
+
+  ram_base = saturn.mem_cntl[1].config[0];
+  ram_mask = saturn.mem_cntl[1].config[1];
+  if (opt_gx)
+    {
+      saturn.mem_cntl[1].config[0] = 0x80000;
+      saturn.mem_cntl[1].config[1] = 0xc0000;
+      dsktop = DSKTOP_GX;
+      dskbot = DSKBOT_GX;
+    }
+  else
+    {
+      saturn.mem_cntl[1].config[0] = 0x70000;
+      saturn.mem_cntl[1].config[1] = 0xf0000;
+      dsktop = DSKTOP_SX;
+      dskbot = DSKBOT_SX;
+    }
+
+  load_addr(&sp, dsktop, 5);
+  load_addr(&end, dskbot, 5);
+
+  n = (end - sp) / 5 - 1;   /* end never matches sp */
+
+  printf("n = %d\n", n);
+
+  //TODO Get the top of the stack in the buffer.
+  //TODO create a window and put the complete stack into it!
+  if (n)
+    {
+      load_addr(&ent, sp, 5);
+      decode_rpl_obj_2(ent, typ, dat);
+      printf("%d %p -> [%s] %s\n", i, ent, typ, dat);
+    }
+
+  for (i = 0; i < n; i++)
+    {
+      load_addr(&ent, sp + (5 * i), 5);
+      decode_rpl_obj_2(ent, typ, dat);
+      printf("%d %p -> [%s] %s\n", i, ent, typ, dat);
+    }
+
+  saturn.mem_cntl[1].config[0] = ram_base;
+  saturn.mem_cntl[1].config[1] = ram_mask;
+
+  return;
+}
+
 static void
 #ifdef __FunctionProto__
 do_stack (int argc, char **argv)
@@ -1493,7 +1661,7 @@ do_stack (argc, argv)
           }
       printf("%5d: %.5lX -> %s\n", se->se_n, se->se_p, buf);
       se = se->se_next;
-    } 
+    }
 
   se = stack;
   while (se)
@@ -1676,6 +1844,7 @@ debug ()
       if (enter_debugger & USER_INTERRUPT)
         if (verbose)
           printf ("%s: user interrupt (SIGINT) ignored\n", progname);
+        exit_x48(1);
       if (enter_debugger & BREAKPOINT_HIT)
         if (verbose)
           printf ("%s: breakpoint hit at 0x%.5lX ignored\n",
@@ -1734,15 +1903,7 @@ debug ()
       /*
        * read a command
        */
-#ifdef HAVE_READLINE
       rl = readline ("x48-debug> ");
-#else
-      if (rl == (char *) 0)
-        rl = (char *)malloc((size_t)80);
-      printf("x48-debug> ");
-      fflush(stdout);
-      rl = read_str(rl, 80, 0);
-#endif
 
       if (rl == (char *) 0)
 	{
@@ -1765,10 +1926,6 @@ debug ()
 	}
       else
 	{
-#ifndef HAVE_READLINE
-          if (rl[strlen(rl) - 1] == '\n')
-            rl[strlen(rl) - 1] = '\0';
-#endif
 	  if (cl)
 	    {
 	      free (cl);
@@ -1833,7 +1990,7 @@ debug ()
   in_debugger = 1;
   ticks = get_t1_t2 ();
   in_debugger = 0;
-  
+
   if (saturn.t2_ctrl & 0x01)
     {
       saturn.timer2 = ticks.t2_ticks;
